"""
CARPI COMMONS
(C) 2018, Raphael "rGunti" Guntersweiler
Licensed under MIT
"""
from logging import Logger, getLogger, info, DEBUG, INFO
from logging.config import fileConfig, dictConfig
from os.path import isfile
from sys import exc_info
from traceback import format_exception

from carpicommons.errors import CarPiExitException

_LOGGER: Logger = None

DEFAULT_CONFIG = dict(
    version=1,
    formatters={
        'f': {
            'format': '[%(asctime)s.%(msecs)03d] [%(name)-15s] %(levelname)-8s %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S'
        }
    },
    handlers={
        'h': {
            'class': 'logging.StreamHandler',
            'formatter': 'f',
            'level': DEBUG
        }
    },
    root={
        'handlers': ['h'],
        'level': INFO
    }
)


def configure_logging(config_files: list,
                      dict_config: dict=None):
    for config_file in config_files:
        if isfile(config_file):
            fileConfig(config_file)
            info("Logging configuration loaded: {}".format(config_file))
            return

    dictConfig(dict_config if dict_config else DEFAULT_CONFIG)
    info("Dictionary configuration loaded"
         if dict_config
         else "Default configuration loaded")


def init_logging(daemon_name: str):
    _LOGGER = getLogger(daemon_name)


def logger(name: str=None) -> Logger:
    return getLogger(name) if name else _LOGGER


def _print_exception(exc_type, exc, traceback):
    out: list = format_exception(exc_type, exc, traceback, limit=64)
    _LOGGER.critical('\n'.join(out))


def log_exception(ex: CarPiExitException):
    _LOGGER.critical("A known error has occurred that forced the daemon to shut down!")
    _LOGGER.critical("The error was:")
    _LOGGER.critical("[%s] %s", ex.exit_code, ex.exc_message)
    _LOGGER.critical("Stack Trace ahead:")
    log_unhandled_exception(False)


def log_unhandled_exception(log_prefix_message: bool=True):
    exc_type, exc, traceback = exc_info()
    if log_prefix_message:
        _LOGGER.critical("An unexpected error has forced the daemon to shut down!")
        _LOGGER.critical("Exception detail:")
    _print_exception(exc_type, exc, traceback)
    del exc_type, exc, traceback



"""
CARPI COMMONS
(C) 2018, Raphael "rGunti" Guntersweiler
Licensed under MIT
"""

"""
CARPI DASH DAEMON
(C) 2018, Raphael "rGunti" Guntersweiler
Licensed under MIT
"""

from datetime import datetime, timedelta


class ExpiringDictionary(dict):
    def __init__(self, timeout=5.0, **kwargs):
        super().__init__(**kwargs)
        self._add_times = {}
        self._timeout: float = timeout

    @property
    def timeout(self) -> float:
        return self._timeout

    @timeout.setter
    def timeout(self, value: float):
        self._timeout = value

    def __getitem__(self, key):
        if self.has_item_expired(key):
            # This will intentionally remove an item from the dict
            # and then raise an a KeyError because expired items
            # should not exist!
            self.__delitem__(key)

        return super().__getitem__(key)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        self._add_times[key] = datetime.now()

    def __delitem__(self, v) -> None:
        super().__delitem__(v)
        del self._add_times[v]

    def get(self, key, default=None):
        try:
            self.__getitem__(key)
        except KeyError:
            return default

    def run_cleanup(self):
        remove_keys = []
        max_dif = timedelta(seconds=self._timeout)
        for key, time in self._add_times:
            if datetime.now() - time > max_dif:
                remove_keys.append(key)

        for key in remove_keys:
            self.__delitem__(key)

    def get_insertion_datetime(self, key) -> datetime:
        """
        Returns the date and time of the insertion of the given key.
        Returns `datetime.min` if the item doesn't exist
        """
        return self._add_times.get(key, datetime.min)

    def get_age(self, key) -> timedelta:
        """
        Returns the age of a given item
        """
        return datetime.now() - self.get_insertion_datetime(key)

    def has_item_expired(self, key) -> bool:
        """
        Returns `true` if an item has been persisted longer than <timeout> seconds ago.
        Returns `true` if an item doesn't exist.
        """
        return self.get_age(key) > timedelta(seconds=self._timeout)



"""
CARPI COMMONS
(C) 2018, Raphael "rGunti" Guntersweiler
Licensed under MIT
"""
import os
from datetime import datetime
from time import time

from setuptools import setup


# ### READING DOCUMENTATION ####################################
def convert_md_to_rst(file: str):
    with open(file, 'r') as f:
        return f.read()


# ### READING VERSION.txt NUMBER ###############################
def read_version():
    with open('VERSION.txt', 'r') as f:
        return f.read().strip()


def read_autobuild_version():
    try:
        with open('VERSION.autobuild.txt', 'r') as f:
            return f.read().strip()
    except FileNotFoundError:
        return None


def get_version():
    autobuild_ver = read_autobuild_version()
    if autobuild_ver:
        return autobuild_ver

    build_time = datetime.fromtimestamp(time())
    base_ver = read_version()

    if 'CI_BUILD_REF_NAME' in os.environ:
        branch = os.environ['CI_BUILD_REF_NAME']
        if branch == "develop":
            # Develop Build
            return '{}.dev{}'.format(base_ver,
                                     build_time.strftime('%y%m%d%H%M%S'))
        elif branch.startswith('release/'):
            # Release Candidate Build
            base_ver = branch.split('/')[1]
            return "{}rc{}".format(base_ver,
                                   build_time.strftime('%y%m%d%H%M%S'))
        else:
            # Release Build
            return os.environ['CI_BUILD_REF_NAME']
    else:
        return '{}.dev{}'.format(base_ver,
                                 build_time.strftime('%y%m%d%H%M%S'))


# ### PRINTING ENVIRONMENT VARIABLES #######################
# for key, val in os.environ.items():
#     print(" -- ENV {} = {}".format(key, val))


# ### DEFINING PROJECT PROPERTIES ##########################
NAME = "carpi-commons"
VERSION = get_version()
DESCRIPTION = open('SHORT_DESCRIPTION.txt', 'r').read()
LONG_DESCRIPTION = convert_md_to_rst('README.md')

AUTHOR = "rGunti"
AUTHOR_EMAIL = "raphael@rgunti.ch"
LICENSE = "MIT"
URL = "https://gitlab.com/rGunti/CarPi-Commons"

PACKAGES = [
    'carpicommons'
]
DEPENDENCIES = [
    'wheel'
]
OPTIONAL_DEPENDENCIES = {}
CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3.6'
]

# ### BUILDING #############################################
print(" >> Building {} Ver. {}".format(NAME, VERSION))
setup(name=NAME,
      version=VERSION,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      license=LICENSE,
      packages=PACKAGES,
      install_requires=DEPENDENCIES,
      extras_require=OPTIONAL_DEPENDENCIES,
      classifiers=CLASSIFIERS,
      url=URL,
      zip_safe=False,
      include_packaged_data=True)



"""
CARPI COMMONS
(C) 2018, Raphael "rGunti" Guntersweiler
Licensed under MIT
"""


class ExitCodes:
    """
    List of valid exit codes;
    Daemon exceptions create exit codes with value 0xFXXX
    """

    OK = 0x0000
    """ Expected when shutting down """

    UNKNOWN = 0xFFFF
    """ Unknown error, unhandled exception or CarPiExitException without error code """

    CONFIG_FAIL = 0xF000
    """ Failure when loading configuration """


class CarPiExitException(BaseException):
    def __init__(self,
                 exit_code=ExitCodes.UNKNOWN):
        self._exit_code = exit_code

    @property
    def exit_code(self):
        return self._exit_code

    @property
    def exc_message(self):
        return u"An unknown exception has occurred. " \
               "The application code will exit with exit code {}".format(self.exit_code)

    def __str__(self) -> str:
        return self.exc_message


class ConfigFileLoadError(CarPiExitException):
    def __init__(self, requested_log_file: str) -> None:
        CarPiExitException.__init__(self,
                                    exit_code=ExitCodes.CONFIG_FAIL)
        self._requested_log_file = requested_log_file

    @property
    def requested_log_file(self) -> str:
        return self._requested_log_file

    @property
    def exc_message(self):
        return u"Failed to load configuration file {}".format(self.requested_log_file)
